/*
 * gMix open source project - https://svs.informatik.uni-hamburg.de/gmix/
 * Copyright (C) 2014  SVS
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "LibSodiumWrapper.h"
#include <sodium.h>

/**
 * This is a JNI wrapper for the libSodium library, which is used for
 * cryptographic operations in the EncDNS Java classes.
 * 
 * Author: Jens Lindemann
 */

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_init
 * Signature: ()V
 */
JNIEXPORT void JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1init
  (JNIEnv *env, jobject jobj) {
    sodium_init();
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_getSKBytes
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1getSKBytes
  (JNIEnv *env, jobject jobj) {
    return crypto_box_PUBLICKEYBYTES;
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_getPKBytes
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1getPKBytes
  (JNIEnv *env, jobject jobj) {
    return crypto_box_SECRETKEYBYTES;
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_getNonceBytes
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1getNonceBytes
  (JNIEnv *env, jobject jobj) {
    return crypto_box_NONCEBYTES;
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_getZeroBytes
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1getZeroBytes
  (JNIEnv *env, jobject jobj) {
    return crypto_box_ZEROBYTES;
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_getBoxZeroBytes
 * Signature: ()I
 */
JNIEXPORT jint JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1getBoxZeroBytes
  (JNIEnv *env, jobject jobj) {
    return crypto_box_BOXZEROBYTES;
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_generateKeys
 * Signature: ([B[B)V
 */
JNIEXPORT void JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1generateKeys
        (JNIEnv *env, jobject jobj, jbyteArray jpk, jbyteArray jsk) {
    unsigned char pk[crypto_box_PUBLICKEYBYTES];
    unsigned char sk[crypto_box_SECRETKEYBYTES];
    
    crypto_box_keypair(pk, sk); // call libsodium's key generation funtion
    
    // Convert public key and send it to Java environment
    (*env)->SetByteArrayRegion(env, jpk, 0, crypto_box_PUBLICKEYBYTES, (jbyte*)pk);
    
    // Convert secret key and send it to Java environment
    (*env)->SetByteArrayRegion(env, jsk, 0, crypto_box_SECRETKEYBYTES, (jbyte*)sk);
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_crypto_box
 * Signature: ([BI[B[B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1crypto_1box
  (JNIEnv *env, jobject jobj, jbyteArray jm, jint mlen, jbyteArray jn, 
        jbyteArray jpk, jbyteArray jsk) {
    char isCopy;
    jbyte* m = (*env)->GetByteArrayElements(env, jm, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jm, m, JNI_ABORT);
    jbyte* n = (*env)->GetByteArrayElements(env, jn, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jn, n, JNI_ABORT);
    jbyte* pk = (*env)->GetByteArrayElements(env, jpk, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jpk, pk, JNI_ABORT);
    jbyte* sk = (*env)->GetByteArrayElements(env, jsk, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jsk, sk, JNI_ABORT);
    
    unsigned char* c = (unsigned char*)calloc(mlen, sizeof(char));
    crypto_box(c, (unsigned char*)m, mlen, (unsigned char*)n, (unsigned char*)pk, (unsigned char*)sk);
    
    jbyteArray out = (*env)->NewByteArray(env, mlen);
    (*env)->SetByteArrayRegion(env, out, 0, mlen, (jbyte*)c);
    
    return out;
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_crypto_box_open
 * Signature: ([BI[B[B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1crypto_1box_1open
  (JNIEnv *env, jobject jobj, jbyteArray jc, jint clen, jbyteArray jn, 
        jbyteArray jpk, jbyteArray jsk) {
    char isCopy;
    jbyte* c = (*env)->GetByteArrayElements(env, jc, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jc, c, JNI_ABORT);
    jbyte* n = (*env)->GetByteArrayElements(env, jn, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jn, n, JNI_ABORT);
    jbyte* pk = (*env)->GetByteArrayElements(env, jpk, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jpk, pk, JNI_ABORT);
    jbyte* sk = (*env)->GetByteArrayElements(env, jsk, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jsk, sk, JNI_ABORT);
    
    unsigned char* m = (unsigned char*)calloc(clen, sizeof(char));
    int ret = crypto_box_open(m, (unsigned char*)c, clen, (unsigned char*)n, (unsigned char*)pk, (unsigned char*)sk);
    if(ret == -1) {
        printf("CryptoBox verfication failed");
    }
    
    jbyteArray out = (*env)->NewByteArray(env, clen);
    (*env)->SetByteArrayRegion(env, out, 0, clen, (jbyte*)m);
    
    return out;
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_crypto_box_beforenm
 * Signature: ([B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1crypto_1box_1beforenm
  (JNIEnv *env, jobject jobj, jbyteArray jpk, jbyteArray jsk) {
        char isCopy;
    jbyte* pk = (*env)->GetByteArrayElements(env, jpk, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jpk, pk, JNI_ABORT);
    jbyte* sk = (*env)->GetByteArrayElements(env, jsk, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jsk, sk, JNI_ABORT);
    
    unsigned char* k = (unsigned char*)calloc(crypto_box_BEFORENMBYTES, sizeof(char));
    
    crypto_box_beforenm(k, (unsigned char*)pk, (unsigned char*)sk);
    
    jbyteArray out = (*env)->NewByteArray(env, crypto_box_BEFORENMBYTES);
    (*env)->SetByteArrayRegion(env, out, 0, crypto_box_BEFORENMBYTES, (jbyte*)k);
    
    return out;
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_crypto_box_afternm
 * Signature: ([BJ[B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1crypto_1box_1afternm
  (JNIEnv *env, jobject jobj, jbyteArray jm, jlong mlen, jbyteArray jn, jbyteArray jk) {
        char isCopy;
    jbyte* m = (*env)->GetByteArrayElements(env, jm, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jm, m, JNI_ABORT);
    jbyte* n = (*env)->GetByteArrayElements(env, jn, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jn, n, JNI_ABORT);
    jbyte* k = (*env)->GetByteArrayElements(env, jk, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jpk, pk, JNI_ABORT);
    
    unsigned char* c = (unsigned char*)calloc(mlen, sizeof(char));
    crypto_box_afternm(c, (unsigned char*)m, mlen, (unsigned char*)n, (unsigned char*)k);
    
    jbyteArray out = (*env)->NewByteArray(env, mlen);
    (*env)->SetByteArrayRegion(env, out, 0, mlen, (jbyte*)c);
    
    return out;
}

/*
 * Class:     com_github_encdns_LibSodiumWrapper
 * Method:    jni_crypto_box_open_afternm
 * Signature: ([BJ[B[B)[B
 */
JNIEXPORT jbyteArray JNICALL Java_com_github_encdns_LibSodiumWrapper_jni_1crypto_1box_1open_1afternm
  (JNIEnv *env, jobject jobj, jbyteArray jc, jlong clen, jbyteArray jn, jbyteArray jk) {
    char isCopy;
    jbyte* c = (*env)->GetByteArrayElements(env, jc, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jc, c, JNI_ABORT);
    jbyte* n = (*env)->GetByteArrayElements(env, jn, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jn, n, JNI_ABORT);
    jbyte* k = (*env)->GetByteArrayElements(env, jk, &isCopy);
    //(*env)->ReleaseByteArrayElements(env, jpk, pk, JNI_ABORT);
    
    unsigned char* m = (unsigned char*)calloc(clen, sizeof(char));
    int ret = crypto_box_open_afternm(m, (unsigned char*)c, clen, (unsigned char*)n, (unsigned char*)k);
    if(ret == -1) {
        printf("CryptoBox verfication failed");
    }
    
    jbyteArray out = (*env)->NewByteArray(env, clen);
    (*env)->SetByteArrayRegion(env, out, 0, clen, (jbyte*)m);
    
    return out;
}