/*
 * gMix open source project - https://svs.informatik.uni-hamburg.de/gmix/
 * Copyright (C) 2014  SVS
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.github.encdns;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.ArrayUtils;

/**
 * This is the EncDNS client proxy (also known as local proxy). It should be
 * executed on the client computer or at least be connected to it via a
 * trustworthy connection. The client proxy will encrypt standard DNS requests
 * sent to it by a stub resolver and pass the encrypted request on to the local
 * recursive nameserver. It will also decrypt the encrypted response received 
 * from the local recursive nameserver and pass the decrypted standard DNS 
 * response on to the stub resolver.
 * 
 * This implementation currently supports UDP only.
 * 
 * @author Jens Lindemann
 */
public class EncDNSClientProxy {
    /** listening port */
    public static final short LISTENING_PORT = 53;
    /** maximum message size */
    public static final int MAX_MSG_SIZE = 65535;
    /** query timeout */
    public static final int TIMEOUT = 5000;
    private int _verbosity;
    private LibSodiumWrapper _libSodium;
    private byte[] _pk;
    private byte[] _sk;
    private byte[] _rPK;
    private byte[] _k;
    private byte[] _magicString;
    private byte[] _remoteNS;
    private InetAddress _localNS;
    private int _localNSPort;
    private DatagramSocket _udpSock53;
    private Random _rnd;
    private boolean _noEncryption;
    
    private static Options _opt;
    private int _currPort;

    /**
     * Creates a new instance of EncDNSClientProxy.
     * @param remotens remote recursive nameserver's EncDNS zone name
     * @param localns locasl recursive nameserver's address
     * @param localnsport local recursive nameserver's port
     * @param rrnskeyPath path to remote recursive nameserver's public key
     * @param noEncryption If true, cryptography will be disabled and unencrypted messages will merely be passed on.
     * @param verbosity verbosity level
     */
    public EncDNSClientProxy(String remotens, InetAddress localns, int localnsport, String rrnskeyPath, boolean noEncryption, int verbosity) {
        _libSodium = new LibSodiumWrapper();
        
        _rnd = new Random();
        
        _verbosity = verbosity;
        
        // Generate new local key pair. We do not need to save this persistently
        // as the public key will always be sent along with a query and using
        // the same key for multiple sessions would make these linkable.
        _pk = new byte[_libSodium.PKBYTES];
        _sk = new byte[_libSodium.SKBYTES];
        _libSodium.jni_generateKeys(_pk, _sk);
        
        // Load remote recursive nameserver's PK from file
        _rPK = EncDNSHelper.readByteArrayFromFile(rrnskeyPath);
        if(_rPK == null) {
            System.err.println("ERROR: Could not load remote recursive "
                    + "nameserver's public key. Exiting.");
            return;
        }
        
        long ktime = System.currentTimeMillis();
        _k = _libSodium.cryptoBoxBeforenm(_rPK, _sk);
        if(_verbosity >= 1) {
            System.out.println("key generation time (ms) = " + (System.currentTimeMillis()-ktime));
        }

        _magicString = new byte[]{0x20, 0x45, 0x5e};
        
        _remoteNS = EncDNSHelper.parseZoneNameString(remotens);
        _localNS = localns;
        _localNSPort = 53;
        _noEncryption = noEncryption;
        _currPort = 1024;
            
        initUDPQueryListener();
    }

    /**
     * @param args the command line arguments (see -h for documentation)
     */
    public static void main(String[] args) {
        try {
            _opt = new Options();
            _opt.addOption("h", "help", false, "prints this message");
            _opt.addOption("d", "disable-encryption", false, "disable encryption");
            _opt.addOption("s", "libsodiumwrap", true, "path to libsodiumWrap library (default: in working directory)");
            _opt.addOption("rk", "rrns-key", true, "path to remote recursive nameserver's public key (default: rrns.pk in working directory on Windows, /etc/encdns-lp/rrns.pk otherwise)");
            _opt.addOption("v", "verbose", false, "enable verbose output");
            
            Option laOpt = new Option("la", "localns", true, "local nameserver's IP address or hostname (REQUIRED)");
            laOpt.setRequired(true);
            _opt.addOption(laOpt);
            
            _opt.addOption("lp", "localnsport", true, "local nameserver's port (default: 53)");
            
            Option raOpt = new Option("ra", "remotens", true, "remote nameserver's zone name (REQUIRED)");
            raOpt.setRequired(true);
            _opt.addOption(raOpt);
            
            Options helpOpts = new Options();
            helpOpts.addOption("h", "help", false, "prints this message");
            
            CommandLineParser parser = new BasicParser();
            
            // We need to check for the help parameter first, as parsing all
            // simultaneously will throw an exception if a required option
            // is missing.
            CommandLine cmd = parser.parse(helpOpts, args, true);
            if (cmd.hasOption("help")) {
                HelpFormatter hf = new HelpFormatter();
                hf.printHelp("EncDNSClientProxy", _opt);
                return;
            }
            
            // Now we can parse all options
            cmd = parser.parse(_opt, args);

            // get parameter values passed by CLI
            String libStr = cmd.getOptionValue("libsodiumwrap");
            String localnsStr = cmd.getOptionValue("localns");
            String localnsportStr = cmd.getOptionValue("localnsport");
            String remotensStr = cmd.getOptionValue("remotens");
            String rrnskeyStr = cmd.getOptionValue("rrns-key");
            
            // set standard values if corresponding CLI parameters are not set
            if(libStr == null) {
                if(EncDNSHelper.osIsWindows()) {
                    libStr = System.getProperty("user.dir") + "\\libsodiumWrap.dll";
                } else { // assume a Unix-like OS
                    libStr = System.getProperty("user.dir") + "/libsodiumWrap.so";
                }
            }
            if(localnsStr == null) localnsStr = "127.0.0.1";
            int localnsport = 53;
            if(localnsportStr != null) {
                localnsport = Integer.parseInt(localnsportStr);
            }
            
            boolean noEncryption = false;
            if(cmd.hasOption("disable-encryption")) {
                noEncryption = true;
            }
            
            int verbosity;
            if(cmd.hasOption("verbose")) {
                verbosity = 1;
            } else {
                verbosity = 0;
            }
            
            if(rrnskeyStr == null) {
                if(EncDNSHelper.osIsWindows()) {
                    rrnskeyStr = System.getProperty("user.dir") + "\\rrns.pk";
                } else { // assume Unix-like OS
                    rrnskeyStr = "/etc/encdns-lp/rrns.pk";
                }
            }

            // load the required C wrapper library
            System.load(libStr);
            
            // resolve the local recursive nameserver's address or IP
            InetAddress localns = InetAddress.getByName(localnsStr);
            
            new EncDNSClientProxy(remotensStr, localns, localnsport, rrnskeyStr, noEncryption, verbosity);
        } catch (MissingOptionException e) {
            // print and error and the help if a required option is missing
             System.out.println(e.getMessage());
             HelpFormatter hf = new HelpFormatter();
             hf.printHelp("EncDNSClientProxy", _opt);
        } catch (ParseException e) {
            System.err.println(e);
        } catch (UnknownHostException e) {
            // print an error if the local recursive nameserver's address cannot
            // be resolved
            System.err.println("local recursive nameserver's address could not be resolved");
        }
    }

    /**
     * Initializes the UDP query listener.
     */
    private void initUDPQueryListener() {
        try {
            _udpSock53 = new DatagramSocket(LISTENING_PORT);
            byte[] rcvbytes = new byte[MAX_MSG_SIZE];
            while (true) {
                DatagramPacket rcvPkt = new DatagramPacket(rcvbytes, rcvbytes.length);
                _udpSock53.receive(rcvPkt);
                
                if(_verbosity >= 1) {
                    System.out.println("Received request from stub");
                }

                // TODO It might be a bit more efficient to use CachedThreadPool.
                
                // Create a new handler thread to avoid blocking the main 
                // control flow.
               ClientMessageHandlerThread msgh = new ClientMessageHandlerThread(rcvPkt);
               new Thread(msgh).start();
            }
        } catch (IOException e) {
            System.err.println("Port busy! Exiting.");
            if(_verbosity >= 1) {
                System.err.println(e);
            }
        }
    }

    /**
     * Encrypts a standard DNS query.
     * @param rcvDNS standard DNS query
     * @return encrypted EncDNS message containing the standard DNS query
     */
    private byte[] encryptQuery(byte[] rcvDNS) {
        // Generate client nonce: 8 byte timer+4 byte random value
        byte[] n = new byte[_libSodium.NONCEBYTES];
        ByteBuffer timeBuffer = ByteBuffer.allocate(8);
        timeBuffer.putLong(System.nanoTime());
        byte[] time = timeBuffer.array();
        byte[] random = new byte[(_libSodium.NONCEBYTES/2)-time.length];
        _rnd.nextBytes(random);
        System.arraycopy(time, 0, n, 0, time.length);
        System.arraycopy(random, 0, n, time.length, random.length);
        
        long ctime = System.currentTimeMillis();
        byte[] cbox = _libSodium.makeCryptoBoxAfternm(rcvDNS, _k, n);
        if(_verbosity >= 1) {
            System.out.println("query encryption time " + (System.currentTimeMillis()-ctime) + " ms");
        }
        
        byte[] encQuery = buildEncQuery(cbox, n);
        return encQuery;
    }

    /**
     * Constructs an EncDNS message containing a specified cryptobox.
     * @param cbox cryptobox to include in message
     * @param nonce nonce used for encryption
     * @return EncDNS message containing the specified cryptobox and nonce
     */
    private byte[] buildEncQuery(byte[] cbox, byte[] nonce) {
        // TODO This uses 65 kB of memory even if the resultig query is a lot
        // smaller ... :-(
        ByteBuffer buf = ByteBuffer.allocateDirect(MAX_MSG_SIZE);

        // ------ CBOX IN HOSTNAME ------
        // --- DNS HEADER ---
        // randomized ID
        byte[] id = new byte[2];
        _rnd.nextBytes(id);
        buf.put(id);
        buf.put((byte) 1); // QR=0 (Query), OpCode=0000 (Query), AA=0, TC=0, RD=1
        buf.put((byte) 0); // RA,Z,AD,CD,RCode=0
        buf.putShort((short) 1); // question count = 1
        buf.putShort((short) 0); // answer count = 0
        buf.putShort((short) 0); // authority count = 0
        buf.putShort((short) 1); // additional count = 1

        // --- DNS QUESTION SECTION ---
        // cbox must be split into labels of length<=63
        int cpos = 0;
        // We have to transmit: magic string, the client's public key, the client's part of the nonce and the cryptobox itself
        byte[] cryptoStuff = ArrayUtils.addAll(_magicString, ArrayUtils.addAll(ArrayUtils.addAll(_pk, Arrays.copyOfRange(nonce, 0, (_libSodium.NONCEBYTES/2))), cbox));
        while (true) {
            if ((cpos + 63) < cryptoStuff.length) {
                // remaining part of cbox > 63 byte -> add first 63 bytes as label
                buf.put((byte)63);
                buf.put(cryptoStuff, cpos, 63);
                cpos += 63;
            } else {
                // remaining part of cbox <= 63 byte -> add as last label
                short remainingBytes = (short) (cryptoStuff.length - cpos);
                buf.put((byte)remainingBytes);
                buf.put(cryptoStuff, cpos, remainingBytes);
                break;
            }
        }

        buf.put(_remoteNS); // put the remote recursive nameserver's EncDNS zone name at the end of the question name
        buf.putShort((short) 16); // TYPE=TXT(16)
        buf.putShort((short) 1); // CLASS=IN(1)

        // --- DNS ADDITIONAL SECTION - EDNS0 OPT PSEUDO-RR ---
        buf.put((byte) 0); // Domain name = root
        buf.putShort((short) 41); // RR TYPE=OPT(41)
        buf.putShort((short) MAX_MSG_SIZE); // sender max UDP payload
        buf.put((byte) 0); // RCode extension=0
        buf.put((byte) 0); // EDNS version=0
        buf.put((byte) 0); // DNSSEC OK=0, Z=0
        buf.put((byte) 0); // Z=0
        buf.putShort((short) 0); // RDATA length

        // convert to byte[]
        byte[] out = new byte[buf.position()];
        buf.position(0);
        buf.get(out);

        return out;
    }

    /**
     * Decrypt a response received from the local recursive nameserver
     * @param encResponse response received from the local recursive nameserver
     * @return decrypted standard DNS message
     */
    private byte[] decryptResponse(byte[] encResponse) {
        if((encResponse[3]&0x0f)!=0) { // if RCODE!=0 (i.e. if an error occurred)
            return null;
        }
        
        // find the position of the question name's end in the message
        int qNameEnd = EncDNSHelper.findQuestionNameEnd(encResponse);
        
        // This is a bit of cheating, but we know that the remote proxy should
        // send only one answer - if there are additional questions or answers
        // in the message, something went wrong (probably at the local recursive
        // nameserver)...
        
        // glue together <character-string>s of split messages
        int txtlen = (((int) encResponse[qNameEnd + 15]) << 8) | (encResponse[qNameEnd+16]&0xff);
        ArrayList<Byte> cryptList = new ArrayList<Byte>();
        for (int mpos = 17; mpos < (17+txtlen);) {
            int cslen = encResponse[qNameEnd+mpos]&0xff; // get length of character-string (and correct sign)
            mpos++;
            for (int i = 0; i < cslen; i++) {
                cryptList.add(encResponse[qNameEnd+mpos]);
                mpos++;
            }
        }

        byte[] cryptoStuff = ArrayUtils.toPrimitive(cryptList.toArray(new Byte[0]));
        byte[] n = Arrays.copyOfRange(cryptoStuff, 0, _libSodium.NONCEBYTES); // nonce
        byte[] cbox = Arrays.copyOfRange(cryptoStuff, _libSodium.NONCEBYTES, cryptoStuff.length); // cryptobox
        
        long time = System.currentTimeMillis();
        // decrypt the message
        byte[] decM = _libSodium.openCryptoBoxAfternm(cbox, _k, n);
        if(_verbosity >= 1) {
            System.out.println("response decryption time " + (System.currentTimeMillis()-time) + " ms");
        }
        return decM;
    }
    
    /**
     * Returns the port to use for sending the next query. This will iterate
     * through all unprivileged (i.e. >=1025) ports.
     * @return port to use for next query
     */
    private synchronized int getCurrPort() {
        if(_currPort==65535) {
            // on overflow, revert to port 1025 (first unprivileged port)
            _currPort = 1025;
        } else {
            _currPort++;
        }
        
        return _currPort;
    }
    
    /**
     * This thread will send a query to the local recursive nameserver, wait for
     * a response and pass the decrypted response on to the stub resolver.
     * 
     * This is encapsulated in a thread to allow parallelized processing of
     * queries.
     */
    private class ClientMessageHandlerThread implements Runnable {
       private DatagramPacket _rcvPkt;
       
       /**
        * Constructor for ClientMessageHandlerThread
        * @param rcvPkt (hopefully) a standard DNS query received from stub resolver
        */
       ClientMessageHandlerThread(DatagramPacket rcvPkt) {
           _rcvPkt = rcvPkt;
       }
       
       public void run() {
           if(_verbosity >= 1) {
                System.out.println("ClientMessageHandler started");
           }
           
           // Copy received standard DNS query into a byte[]
           byte[] rcvDNS = new byte[_rcvPkt.getLength()];
           System.arraycopy(_rcvPkt.getData(), _rcvPkt.getOffset(), rcvDNS, 0,
                   _rcvPkt.getLength());
           byte[] encMsg;
           if(!_noEncryption) {
               // If encryption is enabled, encrypt message
                encMsg = encryptQuery(rcvDNS);
           } else {
               // Else just use the unencrypted message
               encMsg = Arrays.copyOf(rcvDNS, rcvDNS.length);
           }
           
           if (encMsg == null) {
               if(_verbosity >= 1) {
                   System.out.println("Received empty response");
               }
               return;
           }

           byte[] ansReply = sendQueryAndListenForReply(encMsg);
           byte[] decReply;

           if (ansReply != null) {
               byte[] qid = Arrays.copyOfRange(encMsg, 0, 2);
               byte[] rid = Arrays.copyOfRange(ansReply, 0, 2);
               // Compare query and response ID
               if (Arrays.equals(qid, rid)) {
                   if(!_noEncryption) {
                       // If encryption enabled, we will have to decrypt the response
                        decReply = decryptResponse(ansReply);
                   } else {
                       // Otherwise we can just use the response as is
                       decReply = Arrays.copyOf(ansReply, ansReply.length);
                   }
               } else {
                   // If the received message is not an EncDNS message or if it
                   // does not match the query, a server failure will be 
                   // reported to the stub resolver and the port is closed.
                   // The correct response might still be received later, but 
                   // keeping the port open would allow an attacker to try 
                   // multiple forged responses for one query whereas closing
                   // the port only gives him a single chance at the cost of
                   // having to re-query if a stray package arrived at the port.
                   decReply = EncDNSHelper.generateServfail(rcvDNS);
               }

               if (decReply == null) {
                   // If the reply did not contain a decryptable answer,
                   // generate a SERVFAIL message to send to the stub
                   decReply = EncDNSHelper.generateServfail(rcvDNS);
               }
           } else {
               // If the response is empty, a server failure will be reported
               // back to the stub resolver.
               decReply = EncDNSHelper.generateServfail(rcvDNS);
           }
           
           // Send unencrypted standard DNS response to stub resolver
           DatagramPacket sendPacket = new DatagramPacket(decReply, decReply.length, _rcvPkt.getAddress(), _rcvPkt.getPort());
           try {
               _udpSock53.send(sendPacket);
           } catch (IOException e) {
               System.err.println("Error when sending response to stub resolver");
               if(_verbosity >= 1) {
                System.err.println(e);
               }
           }
       }
       
       /**
        * Sends a query to the local recursive nameserver and waits for a
        * response.
        * @param query query to be sent
        * @return response received
        */
       private byte[] sendQueryAndListenForReply(byte[] query) {
           try {
               // Open a socket on a new port
               DatagramSocket udpSock = null;
               while (udpSock == null) {
                   try {
                       int port = getCurrPort();
                       udpSock = new DatagramSocket(port);
                   } catch (SocketException e) {
                       // do nothing as this is being handled by the while loop
                   }
               }
               
               // Send the encrypted query to the local recursive nameserver
               DatagramPacket sendPacket = new DatagramPacket(query,
                       query.length, _localNS, _localNSPort);
               udpSock.send(sendPacket);
               
               // Wait for a response
               byte[] rcvbytes = new byte[MAX_MSG_SIZE];
               DatagramPacket rcvPkt = new DatagramPacket(rcvbytes, rcvbytes.length);
               udpSock.setSoTimeout(TIMEOUT); // Make sure we do not wait for ages...
               udpSock.receive(rcvPkt);
               udpSock.close();
               byte[] rcvDNS = new byte[rcvPkt.getLength()];
               System.arraycopy(rcvPkt.getData(), rcvPkt.getOffset(), rcvDNS, 0, rcvPkt.getLength());
               if(_verbosity >= 1) {
                    System.out.println("Received reply from authoritative nameserver");
               }
               return rcvDNS;
           } catch(SocketTimeoutException e) {
               if(_verbosity >= 1) {
                System.err.println("Query timed out.");
               }
               return null;
           } catch (IOException e) {
               System.err.println("Error when sending query");
               if(_verbosity >= 1) {
                System.err.println(e);
               }
               return null;
           }
       }
   }
}