/*
 * gMix open source project - https://svs.informatik.uni-hamburg.de/gmix/
 * Copyright (C) 2014  SVS
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.github.encdns;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.MissingOptionException;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.ArrayUtils;

/**
 * This is the EncDNS server proxy (also known as remote proxy). It should be
 * executed on the same computer as the remote recursive nameserver or at least 
 * be connected to it via a trustworthy connection. The server proxy will 
 * decrypt encrypted EncDNS queries it receives and pass the decrypted
 * standard DNS query on to the remote recursive nameserver. It will also
 * encrypt the standard DNS responses received from the remote recursive
 * nameserver and pass the EncDNS response on to the local recursive nameserver.
 * 
 * @author Jens Lindemann, Karl-Peter Fuchs
 */
public class EncDNSServerProxy {
    /** The port on which the server proxy is listening for requests. */
    public static final short LISTENING_PORT = 53;
    /** maximum size of an EncDNS message */
    public static final int MAX_MSG_SIZE = 65535; // = maximum DNS msg length
    /** timeout for queries sent to the remote recursive nameserver */
    private int _timeout;
    /** maximum number of threads */
    private int _maxThreads;
    /** size of cache for intermediate shared secrets */
    public final int CACHESIZE;
    
    private LibSodiumWrapper _libSodium;
    
    private byte[] _pk;
    private byte[] _sk;
    
    private byte[] _magicString;
    
    DatagramSocket _udpSock53;
    private Random _rnd;
    private Map<EncDNSKey, byte[]> _cache;
    private boolean _noEncryption;
    
    private InetAddress _targetNSAddress;
    private int _targetNSPort;
    private byte[] _zoneName;
    
    private static Options _opt;
    
    private int _verbosity;

    private static boolean displayByteThroughput;
    private static boolean displayPacketThroughput;
    private static boolean displayThreadStatus;
    private static int displayThreadInterval;
    private static AtomicInteger decryptingThreads = new AtomicInteger(0);
    private static AtomicInteger resolvingThreads = new AtomicInteger(0);
    private static AtomicInteger encryptingThreads = new AtomicInteger(0);
    private static AtomicInteger sendingThreads = new AtomicInteger(0);
    private static AtomicInteger idleThreads = new AtomicInteger(0);
    private static int threads;
    
    /**
     * Creates a new instance of EncDNSServerProxy.new file
     * @param zoneurl EncDNS zone name (The IP this proxy is listening on must be listed as an authoritative nameserver IP for this zone in the parent zone)
     * @param nsaddr remote recursive nameserver's address
     * @param port remote recursive nameserver's port
     * @param cachesize size of cache for intermediate shared secrets
     * @param pkPath path to public key file
     * @param skPath path to secret key file
     * @param noEncryption If true, disables encryption. If false, encryption is enabled.
     * @param verbosity verbosity level
     */
    public EncDNSServerProxy(String zoneurl, InetAddress nsaddr, int port, int cachesize, int timeout, int maxThreads, String pkPath, String skPath, boolean noEncryption, int verbosity,  boolean displayThreadStatusBool, int displayThreadStatusInt, boolean displayThroughputBool) {
        _libSodium = new LibSodiumWrapper();
        CACHESIZE = cachesize;
        _zoneName = EncDNSHelper.parseZoneNameString(zoneurl);

        _magicString = new byte[]{0x20, 0x45, 0x5e};
        _rnd = new Random();
        _cache = Collections.synchronizedMap(new SharedSecretCache(CACHESIZE));
        _timeout = timeout;
        _maxThreads = maxThreads;
        threads = _maxThreads;
        
        _pk = EncDNSHelper.readByteArrayFromFile(pkPath);
        _sk = EncDNSHelper.readByteArrayFromFile(skPath);
        
        _targetNSAddress = nsaddr;
        _targetNSPort = port;
        _noEncryption = noEncryption;
        _verbosity = verbosity;

        displayThreadInterval = displayThreadStatusInt;
        displayThreadStatus = displayThreadStatusBool;
        displayByteThroughput = displayThroughputBool;
        displayPacketThroughput = displayThroughputBool;
        initUDPQueryListener();
        
        if (displayThreadStatus) {
        	new Thread(
        			new Runnable() {
        				public void run() {
        					while (true) {
        						try {
									Thread.sleep(displayThreadInterval);
								} catch (InterruptedException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
									continue;
								}
        						String threadStatus = "\nThreadStatus (of "+threads +" threads):\n";
        						threadStatus += "  idle threads: " +idleThreads.get() +"\n";
        						threadStatus += "  decrypting threads: " +decryptingThreads.get() +"\n";
        						threadStatus += "  resolvingThreads threads: " +resolvingThreads.get() +"\n";
        						threadStatus += "  encryptingThreads threads: " +encryptingThreads.get() +"\n";
        						System.out.println(threadStatus); 
        					} 
        				}
        			}
        		).start(); 
        	}
        
        if (displayThroughputBool) {
        	StatisticsRecorder.init(displayThroughputBool, displayThroughputBool, displayThreadInterval, threads);
        }

    }

    /**
     * Main method
     * @param args the command line arguments (see -h for documentation)
     */
    public static void main(String[] args) {
        try {
            _opt = new Options();
            _opt.addOption("h", "help", false, "prints this message");
            _opt.addOption("s", "libsodiumwrap", true, "path to libsodiumWrap library (default: in working directory");
            _opt.addOption("a", "nameserver", true, "nameserver's IP address or hostname (default: 127.0.0.1)");
            _opt.addOption("p", "port", true, "nameserver's port (default: 53)");
            _opt.addOption("pk", "public-key", true, "path to public key (default: working directory\\encdns.pk for Windows, /etc/encdns-sp/encdns.pk else)");
            _opt.addOption("sk", "secret-key", true, "path to secret key (default: working directory\\encdns.sk for Windows, /etc/encdns-sp/encdns.sk else)");
            _opt.addOption("c", "cache", true, "shared secret cache size (default: 100)");
            _opt.addOption("t", "timeout", true, "timeout for queries to authoritative nameservers in ms (default: 5000)");
            _opt.addOption("mt", "max-threads", true, "maximum number of threads (default: 100)");
            _opt.addOption("d", "disable-encryption", false, "disable encryption");
           _opt.addOption("v", "verbose", false, "enable verbose output");
           _opt.addOption("dts", "displayThreadStatus", false, "display info about threads (idle, encrypting...)");
           _opt.addOption("dtsi", "displayThreadStatusInterval", true, "interval to display info about threads in ms (default: 1000)");
           _opt.addOption("dt", "displayThroughputStatistics", false, "display info about current throughput");
           
           Option zOpt = new Option("z", "zonename", true, "zone name (REQUIRED)");
            zOpt.setRequired(true);
            _opt.addOption(zOpt);

            Options helpOpts = new Options();
            helpOpts.addOption("h", "help", false, "prints this message");
            
            CommandLineParser parser = new BasicParser();
            
            // We need to check for the help parameter first, as parsing all
            // simultaneously will throw an exception if a required option
            // is missing.
            CommandLine cmd = parser.parse(helpOpts, args, true);
            if (cmd.hasOption("help")) {
                HelpFormatter hf = new HelpFormatter();
                hf.printHelp("EncDNSServerProxy", _opt);
                return;
            }
            
            // Now we can parse all options
            cmd = parser.parse(_opt, args);

            // get parameter values passed by CLI
            String libStr = cmd.getOptionValue("libsodiumwrap");
            String nsStr = cmd.getOptionValue("nameserver");
            String portStr = cmd.getOptionValue("port");
            String zoneStr = cmd.getOptionValue("zonename");
            String cacheStr = cmd.getOptionValue("cache");
            String timeoutStr = cmd.getOptionValue("timeout");
            String threadsStr = cmd.getOptionValue("max-threads");
            String pkStr = cmd.getOptionValue("public-key");
            String skStr = cmd.getOptionValue("secret-key");
            String displayThreadStatusIntStr = cmd.getOptionValue("displayThreadStatusInterval");
            
            // set standard values if corresponding CLI parameters are not set
            if(libStr == null) {
                if(EncDNSHelper.osIsWindows()) {
                    libStr = System.getProperty("user.dir") + "\\libsodiumWrap.dll";
                } else { // assume a Unix-like OS
                    libStr = System.getProperty("user.dir") + "/libsodiumWrap.so";
                }
            }
            
            if(nsStr == null) nsStr = "127.0.0.1";
            int port = 53;
            if (portStr != null) {
                port = Integer.parseInt(portStr);
            }
            int cache = 100;
            if(cacheStr != null) {
                cache = Integer.parseInt(cacheStr);
            }
            int timeout = 5000;
            if(timeoutStr != null) {
                timeout = Integer.parseInt(timeoutStr);
            }
            int maxThreads = 100;
            if(threadsStr != null) {
                maxThreads = Integer.parseInt(threadsStr);
            }
            boolean noEncryption = false;
            if(cmd.hasOption("disable-encryption")) {
                noEncryption = true;
            }
            
            if(pkStr == null) {
                if(EncDNSHelper.osIsWindows()) {
                    pkStr = System.getProperty("user.dir") + "\\encdns.pk";
                } else { // assume a Unix-like OS
                    pkStr = "/etc/encdns-sp/encdns.pk";
                }
            }
            
            if(skStr == null) {
                if(EncDNSHelper.osIsWindows()) {
                    skStr = System.getProperty("user.dir") + "\\encdns.sk";
                } else { // assume a Unix-like OS
                    skStr = "/etc/encdns-sp/encdns.sk";
                }
            }
            
            int verbosity;
            if(cmd.hasOption("verbose")) {
                verbosity = 1;
            } else {
                verbosity = 0;
            }
            
            boolean displayThreadStatusBool;
            if(cmd.hasOption("displayThreadStatus")) {
            	displayThreadStatusBool = true;
            } else {
            	displayThreadStatusBool = false;
            }
            
            int displayThreadStatusInt = 1000;
            if(displayThreadStatusIntStr != null) {
                cache = Integer.parseInt(displayThreadStatusIntStr);
            }
            
            boolean displayThroughputStatisticsBool;
            if(cmd.hasOption("displayThroughputStatistics")) {
            	displayThroughputStatisticsBool = true;
            } else {
            	displayThroughputStatisticsBool = false;
            }

            // load the required C wrapper library
            System.load(libStr);
            
            // resolve the remote recursive nameserver's address or IP
            InetAddress nsaddr = InetAddress.getByName(nsStr);
            
            new EncDNSServerProxy(zoneStr, nsaddr, port, cache, timeout, maxThreads, pkStr, skStr, noEncryption, verbosity, displayThreadStatusBool, displayThreadStatusInt, displayThroughputStatisticsBool);
        } catch (MissingOptionException e) {
            // print an error and the help if a required option is missing
            System.out.println(e.getMessage());
            HelpFormatter hf = new HelpFormatter();
            hf.printHelp("EncDNSServerProxy", _opt);
        } catch (ParseException e) {
            System.err.println("Error when parsing command line options:");
            System.err.println(e);
        } catch (UnknownHostException e) {
            // print an error if the remote recursive nameserver's address 
            // cannot be resolved
            System.err.println("Could not resolve remote recursive nameserver's address");
        }
    }

    /**
     * Initializes the UDP query listener.
     */
    private void initUDPQueryListener() {
        try {
            _udpSock53 = new DatagramSocket(LISTENING_PORT);
            //_udpSock53.setReceiveBufferSize(10000);
            
         //while (true) {
            for (int i=0; i<_maxThreads; i++) {
            	new Thread(new ServerMessageHandlerThread(_udpSock53)).start();
            } 
               
                //ExecutorService tp = Executors.newFixedThreadPool(_maxThreads);
               // LinkedList<Future> tl = new LinkedList<Future>();
                // Create a new handler thread to avoid blocking the main 
                // control flow.
                //ServerMessageHandlerThread msgh = new ServerMessageHandlerThread(rcvPkt);
                //tl.add(tp.submit(msgh));
                //if(tl.size() > _maxThreads) {
                //    tl.pop().cancel(true);
                //}
                //new Thread(msgh).start();
            //}
        } catch (IOException e) {
            System.err.println("UDP socket error:");
            System.err.println(e);
        }
    }
    
    /**
     * Encrypts a standard DNS response.
     * @param encQuery EncDNS query
     * @param qNameEnd position of the question name's end
     * @param response standard DNS response to encrypt
     * @param k intermediate shared secret to use for encryption
     * @return the encrypted EndDNS response containing the specified standard DNS response
     */
    private byte[] encryptResponse(byte[] encQuery, int qNameEnd, byte[] response, byte[] k) {
        // TODO This uses 65 kB of memory even if the resultig message is a lot
        // smaller ... :-(
        ByteBuffer buf = ByteBuffer.allocateDirect(MAX_MSG_SIZE);

        // --- DNS HEADER ---
        // ID must match the query ID
        byte[] id = Arrays.copyOfRange(encQuery, 0, 2);
        buf.put(id);
        
        buf.put((byte) 0x81); // QR=1 (Response), OpCode=0000 (Standard), AA=0, TC=0, RD=1
        buf.put((byte) 0); // RA=0, Z,AD,CD=0, RCode=0000 (no error)
        buf.putShort((short) 1); // question count = 1
        buf.putShort((short) 1); // answer count = 1
        buf.putShort((short) 0); // authority count = 0
        buf.putShort((short) 0); // additional count = 0

        // --- DNS QUESTION SECTION ---
        // copy the question from the query
        buf.put(Arrays.copyOfRange(encQuery, 12, qNameEnd + 5));

        // --- DNS ANSWER SECTION ---
        buf.put((byte) 0xc0); // c00c is the address of the question sections name
        buf.put((byte) 0x0c);
        buf.putShort((short) 16); // TYPE=TXT
        buf.putShort((short) 1); // CLASS = IN
        buf.putInt(0); // TTL = 0

        // copy client nonce
        byte[] nonce = new byte[_libSodium.NONCEBYTES];
        System.arraycopy(encQuery, 13 + _magicString.length + _libSodium.PKBYTES, nonce, 0, _libSodium.NONCEBYTES / 2);
        
        // Generate and add server nonce: 8 byte timer+4 byte random value
        byte[] sn = new byte[_libSodium.NONCEBYTES/2];
        ByteBuffer timeBuffer = ByteBuffer.allocate(8);
        timeBuffer.putLong(System.nanoTime());
        byte[] time = timeBuffer.array();
        byte[] random = new byte[(_libSodium.NONCEBYTES/2)-time.length];
        _rnd.nextBytes(random);
        System.arraycopy(time, 0, sn, 0, time.length);
        System.arraycopy(random, 0, sn, time.length, random.length);
        
        System.arraycopy(sn, 0, nonce, _libSodium.NONCEBYTES/2, _libSodium.NONCEBYTES/2);
        
        // encrypt the message
        byte[] cbox = _libSodium.makeCryptoBoxAfternm(response, k, nonce);
        
        byte[] cryptoStuff = ArrayUtils.addAll(nonce, cbox);
        
        // The contents of a TXT RR must be split into <character-string>s
        // of a length of less than 256 bytes. Each of these has a length byte,
        // so we'll need space for those in the RR.
        short txtlen = (short) (cryptoStuff.length + (cryptoStuff.length / 255) + 1);
        buf.putShort(txtlen);

        // Put nonce and cbox into TXT RR.
        int cpos = 0;
        while (true) {
            if ((cpos + 255) < cryptoStuff.length) {
                // remaining part > 255 byte -> add 255-byte <character-string>
                buf.put((byte) 255);
                buf.put(cryptoStuff, cpos, 255);
                cpos += 255;
            } else {
                // remaining part of cbox <= 63 byte -> add remaining part
                short remainingBytes = (short) (cryptoStuff.length - cpos);
                buf.put((byte) remainingBytes);
                buf.put(cryptoStuff, cpos, remainingBytes);
                break;
            }
        }

        // TODO We may want to copy the payload size from the request. As the
        // value is fixed in the current client implementation, we'll just
        // assume the fixed value here. Also, there is currently no check
        // to prevent messages exceeding this size to be sent.
        // --- DNS ADDITIONAL SECTION - OPT PSEUDO-RR ---
        buf.put((byte) 0); // Domain name = root
        buf.putShort((short) 41); // RR TYPE=OPT(41)
        buf.putShort((short) MAX_MSG_SIZE); // sender max UDP payload=65535
        buf.put((byte) 0); // RCode extension=0
        buf.put((byte) 0); // EDNS version=0
        buf.put((byte) 0); // DNSSEC OK=0, Z=0
        buf.put((byte) 0); // Z=0
        buf.putShort((short) 0); // RDATA length

        // convert to byte[]
        byte[] out = new byte[buf.position()];
        buf.position(0);
        buf.get(out);

        return out;
    }
    
    /**
     * This thread will send a query to the remote recursive nameserver, wait 
     * for a response and pass the decrypted response on to the local recursive
     * nameserver.
     * 
     * This is encapsulated in a thread to allow parallelized processing of
     * queries.
     */
    private class ServerMessageHandlerThread implements Runnable {
        //private DatagramPacket _rcvPkt;
    	private DatagramSocket _udpSock53;
        private byte[] _k;
        private int threadId;

        /**
         * Constructor for ServerMessageHandlerThread
         * @param rcvPkt (hopefully) an EncDNS query received from a loccal
         * recursive nameserver
         */
        ServerMessageHandlerThread(DatagramSocket _udpSock53/*DatagramPacket rcvPkt*/) {
        	this._udpSock53 = _udpSock53;
        	this.threadId = StatisticsRecorder.getThreadId();
        }

        public void run() {
        	if(_verbosity >= 1) {
                System.out.println("ServerMessageHandler started");
            }
        	while (true) {
        		byte[] rcvPkt = new byte[MAX_MSG_SIZE];
        		DatagramPacket _rcvPkt = new DatagramPacket(rcvPkt, rcvPkt.length);
        		if (displayThreadStatus) {
        			sendingThreads.decrementAndGet();
        			idleThreads.incrementAndGet();
            	}
        		// reiceive:
                synchronized (_udpSock53) {
                	try {
						_udpSock53.receive(_rcvPkt);
						if (displayPacketThroughput || displayByteThroughput)
							StatisticsRecorder.addRequestThroughputRecord(_rcvPkt.getLength(), threadId);
					} catch (IOException e) {
						if(_verbosity >= 1) {
							System.err.println("Failed to receive message:");
                            e.printStackTrace();
			            }
						continue;
					}
                }
                if(_verbosity >= 1) {
                    System.out.println("Received message");
                }
                if (displayThreadStatus) {
                	idleThreads.decrementAndGet();
                	decryptingThreads.incrementAndGet();
            	}
             // Copy received query into a byte[]
                byte[] rcvDNS = new byte[_rcvPkt.getLength()];
                System.arraycopy(_rcvPkt.getData(), _rcvPkt.getOffset(), rcvDNS, 0,
                        _rcvPkt.getLength());
                
                int qNameEnd = EncDNSHelper.findQuestionNameEnd(rcvDNS);
                
                byte[] decMsg;
                if(!_noEncryption) {
                    // If encryption is enabled, decrypt the message
                    decMsg = decryptQuery(rcvDNS, qNameEnd);
                } else {
                    // else just pass the unencrypted message on
                    decMsg = Arrays.copyOf(rcvDNS, rcvDNS.length);
                }
                if (decMsg == null) {
                    // If the decrypted message is null, decryption failed (probably due to a manipulated message)
                    if(_verbosity >= 1) {
                        System.out.println("Decryption of EncDNS message failed.");
                    }
                    return;
                }

                if (displayThreadStatus) {
                	decryptingThreads.decrementAndGet();
                	resolvingThreads.incrementAndGet();
            	}
                byte[] ansReply = sendQueryAndListenForReply(decMsg);

                if (displayThreadStatus) {
                	resolvingThreads.decrementAndGet();
                	encryptingThreads.incrementAndGet();
            	}
                byte[] qID = Arrays.copyOfRange(decMsg, 0, 2);
                byte[] rID = Arrays.copyOfRange(ansReply, 0, 2);
                byte[] encReply;
                // Check for empty responses and compare query and response IDs
                if (ansReply != null && Arrays.equals(qID, rID)) {
                    if(!_noEncryption) {
                        // If encryption is enabled, encrypt the response
                        encReply = encryptResponse(rcvDNS, qNameEnd, ansReply, _k);
                    } else {
                        // Else just pass the unencrypted response on
                        encReply = Arrays.copyOf(ansReply, ansReply.length);
                    }
                } else {
                    // If we received an empty response or the response ID does not
                    // match the query ID, something went wrong -> generate a
                    // SERVFAIL message
                    if(!_noEncryption) {
                        encReply = encryptResponse(rcvDNS, qNameEnd, EncDNSHelper.generateServfail(rcvDNS), _k);
                    } else {
                        encReply = EncDNSHelper.generateServfail(rcvDNS);
                    }
                }
                
                if (displayThreadStatus) {
                	encryptingThreads.decrementAndGet();
                	sendingThreads.incrementAndGet();
        		}
                
             // Send response to local recursive nameserver
                DatagramPacket sendPacket = new DatagramPacket(encReply, encReply.length, _rcvPkt.getAddress(), _rcvPkt.getPort());
                
                //synchronized (_udpSock53) {
                	try {
                		if (displayPacketThroughput || displayByteThroughput)
							StatisticsRecorder.addReplyThroughputRecord(encReply.length, threadId);
                        _udpSock53.send(sendPacket);
                    } catch (IOException e) {
                        if(_verbosity >= 1) {
                            System.err.println("Failed to send message:");
                            System.err.println(e);
                        }
                    }
                //}
        	} 
        }

        /**
         * Sends a query to the remote recursive nameserver and waits for a
         * response.
         * @param query query to be sent
         * @return response received
         */
        private byte[] sendQueryAndListenForReply(byte[] query) {
            try {
                // Open a socket on a new port
                DatagramSocket udpSock = null;
                while (udpSock == null) {
                    try {
                        int port = 1024 + _rnd.nextInt(64512);
                        udpSock = new DatagramSocket(port);
                    } catch (SocketException e) {
                        // do nothing as this is being handled by the while loop
                    }
                }

                // Send the decrypted query to the remote recursive nameserver
                DatagramPacket sendPacket = new DatagramPacket(query,
                        query.length, _targetNSAddress, _targetNSPort);
                udpSock.send(sendPacket);
                
                // Wait for a response
                byte[] rcvbytes = new byte[MAX_MSG_SIZE];
                DatagramPacket rcvPkt = new DatagramPacket(rcvbytes, rcvbytes.length);
                udpSock.setSoTimeout(_timeout); // Make sure we do not wait for ages...
                udpSock.receive(rcvPkt);
                udpSock.close();
                byte[] rcvDNS = new byte[rcvPkt.getLength()];
                System.arraycopy(rcvPkt.getData(), rcvPkt.getOffset(), rcvDNS, 0, rcvPkt.getLength());
                if(_verbosity >= 1) {
                    System.out.println("Received reply from aNS");
                }
                return rcvDNS;
            } catch (SocketTimeoutException e) {
                if(_verbosity >= 1) {
                    System.err.println("Query timed out!");
                }
                return null;
            } catch (IOException e) {
                System.err.println("IO error when querying nameserver");
                if(_verbosity >= 1) {
                    System.err.println(e);
                }
                return null;
            }
        }

        /**
         * Decrypts a query.
         * @param rcv encrypted query
         * @param qNameEnd position of the question name's end
         * @return the decrypted query
         */
        private byte[] decryptQuery(byte[] rcv, int qNameEnd) {
            // expected start position of EncDNS zone name
            int msgZoneNameStart = qNameEnd + 1 - _zoneName.length;
            // check whether the EncDNS zone name in the query is correct
            byte[] rcvZoneName = Arrays.copyOfRange(rcv, msgZoneNameStart, qNameEnd + 1);
            if ((msgZoneNameStart > rcv.length) || (!Arrays.equals(_zoneName, rcvZoneName))) {
                if(_verbosity >= 1) {
                    System.err.println("DNS message was not destined for this server!");
                }
                return null;
            }

            // check for the magic string
            if (rcv[12] < _magicString.length || (!Arrays.equals(_magicString, Arrays.copyOfRange(rcv, 13, 13 + _magicString.length)))) {
                if(_verbosity >= 1) {
                    System.err.println("Message not in EncDNS format!");
                }
                return null;
            }

            // copy the crypto information from the query
            ArrayList<Byte> cryptList = new ArrayList<Byte>();
            for (int mpos = 12; mpos < msgZoneNameStart;) {
                int llen = rcv[mpos];
                mpos++;
                for (int i = 0; i < llen; i++) {
                    cryptList.add(rcv[mpos]);
                    mpos++;
                }
            }
            byte[] cryptoStuff = ArrayUtils.toPrimitive(cryptList.toArray(new Byte[0]));
            
            // split cryptographic information
            int pkAndNonceEnd = _magicString.length + _libSodium.PKBYTES + (_libSodium.NONCEBYTES / 2);
            byte[] rpk = Arrays.copyOfRange(cryptoStuff, _magicString.length, _magicString.length + _libSodium.PKBYTES);
            byte[] n = new byte[_libSodium.NONCEBYTES];
            System.arraycopy(cryptoStuff, _magicString.length + _libSodium.PKBYTES, n, 0, _libSodium.NONCEBYTES / 2);
            byte[] cbox = Arrays.copyOfRange(cryptoStuff, pkAndNonceEnd, cryptoStuff.length);

            // check whether we've got the corresponding intermediate shared
            // secret in the cache
            EncDNSKey rpkobj = new EncDNSKey(rpk);
            _k = _cache.get(rpkobj);
            if (_k == null) {
                if(_verbosity >= 1) {
                    System.out.println("Key not found in cache. Calculating intermediate shared secret...");
                }
                _k = _libSodium.cryptoBoxBeforenm(rpk, _sk);
                _cache.put(rpkobj, _k);
            } else {
                if(_verbosity >= 1) {
                    System.out.println("Key found in cache.");
                }
            }

            // decrypt and return the query
            byte[] decCbox = _libSodium.openCryptoBoxAfternm(cbox, _k, n);
            return decCbox;
        }
    }
    
}